(* Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the vfs main unit.
 *
 * The Initial Developer of the Original Code is
 * Noeska Software.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 *  M van der Honing
 *
 *)

unit vfs;

interface

uses classes, sysutils;

type
  TBlock = record
    previous: int64;
    next: int64;
    data: pointer;
  end;
  PBlock = ^TBlock;

  TBuffer = array of byte; //used for copying pointer/buffers

  TFileRecord = packed record
    filename: string[255];
    filesize: int64;
    startblock: int64;
  end;
  PFileRecord = ^TFileRecord;

  TBlockList= array of integer;

  //TODO: implement fmcreate / fmopenread etc modes
  TvfsFileStream = class(TStream)
  private
    FSize: Integer;
    FPosition: Int64;
    FFileRecord: TFileRecord; //name / begin block / stream (file) size
    FBlockList: TBlockList; //blocks used by stream in order
    FDataFileP: TObject; //pointer to virtual filesystem object
    FFirstBlock: int64;
  protected
    function GetSize: Int64; override;
    procedure SetSize(NewSize: LongInt); override;
  public
    constructor Create(aName: string; aVFSFileSystem: TObject);
    function Read(var Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    property Position: Int64 read FPosition write FPosition;
  end;

  TVirtualFileSystem = class
  private
    FDataFile: TFileStream;
    FDataFileName: string;
    FBlockSize: integer;
    FDir: TvfsFileStream;


    //TODO: Special ReadBlock/WriteBlock with begin and endposition
    //TODO: optimize ReadBlock/WriteBlock without size
    function ReadBlock(ablockid: int64; pBuf: pointer): int64; overload;
    function ReadBlock(ablockid: int64; pBuf: pointer; beginpos: integer; endpos: integer): int64; overload;
    function ReadBlock(ablockid: int64; var buffer; beginpos: integer; endpos: integer): int64; overload;

    procedure WriteBlock(ablockid, anextblockid, aprevblockid: int64; pBuf: pointer); overload;
    procedure WriteBlock(ablockid, anextblockid, aprevblockid: int64; pBuf: pointer; beginpos: integer; endpos: integer); overload;
    procedure WriteBlock(ablockid, anextblockid, aprevblockid: int64; const buffer; beginpos: integer; endpos: integer); overload;

    function NextFreeBlock(): int64;
    function SecondNextFreeBlock(): int64;
    
    procedure WriteHeader();
    procedure WriteFirstDirectoryEntry();
    procedure GetBlockId(var existingblock: Boolean; iEndBlock: Integer; blocklistsize: Integer; var BlockList: TBlockList; i: Integer; var b: Int64; var nb: Int64);

  protected
  public
    constructor Create();
    destructor Destroy(); override;
    procedure New(); overload;
    procedure New(aBlockSize: integer); overload;
    procedure Open();
    procedure Close();

    //--------------------------------------------------------------------------

    procedure GetFileList(List: TStrings);
    function FindFile(aname: string; var afilerecord: TFileRecord): int64;
    procedure AddFile(aname: string; astartblockid: int64; asize: int64);
    procedure ChangeFile(aname: string; afilerecord: TFileRecord);

    //--------------------------------------------------------------------------

    function ReadData(aname: string; pBuf: pointer): integer;
    procedure WriteData(aname: string; pBuf: pointer; ASize: integer);

    function  ReadStream(Section, Ident: string; Stream: TStream): Integer;
    function  WriteStream(Section, Ident: string; Stream: TStream): Integer;

    //--------------------------------------------------------------------------

    function GetBlockList(afile: TFileRecord): TBlockList;
    function ReadVirtualFileStreamBuffer(var buffer; Position, Count: Integer; BlockList: TBlockList ): LongInt;
    function WriteVirtualFileStreamBuffer(const buffer; Position, Count: Integer; var BlockList: TBlockList ): LongInt;

    //--------------------------------------------------------------------------

    property FileName: string read FDataFileName write FDataFileName;
    property BlockSize: integer read FBlockSize;
  end;

implementation

//TvfsFileStream ---------------------------------------------------------------

constructor TvfsFileStream.Create(aname: string; aVFSFileSystem: TObject);
var
  FileRecordPos: int64;
begin
  inherited Create();

  //determine if aVFSFileSystem is a TVirtualFileSystem
  if aVFSFileSystem <> nil then
    if aVFSFileSystem is TVirtualFileSystem then
    begin
      self.FDataFileP := aVFSFileSystem;

      //Does aname exist?
      FileRecordPos := TVirtualFileSystem(self.FDataFileP).FindFile(aname,FFileRecord);

      if FFileRecord.startblock > 0 then
      begin
        //TODO: overwrite current file (e.g. delete it dependend on fmode)
        FSize:=FFileRecord.filesize;
        FFirstBlock := FFileRecord.startblock;
      end
      else
      begin
        //new file
        FFirstBlock := TVirtualFileSystem(self.FDataFileP).NextFreeBlock();
        TVirtualFileSystem(self.FDataFileP).AddFile(aname,FFirstBlock , 0);
        FSize:=0;
      end;

      FPosition:=0; //go to start of virtual file

      //determine what blocks are used by virtual file (if any)
      FBlockList:=TVirtualFileSystem(self.FDataFileP).GetBlockList(FFileRecord);

      //some more for making a new virtual file
      if FFilerecord.filename = 'Not Found' then
      begin
        FFileRecord.filename := aname;
        FFileRecord.startblock :=TVirtualFileSystem(self.FDataFileP).NextFreeBlock();
      end;
      FPosition:=0; //go back to start of virtual file
    end
    else
      Raise Exception.Create('aVFSFileSystem must be a TVirtualFileSystem');

end;

function TvfsFileStream.GetSize(): Int64;
begin
  result :=FSize;
end;

procedure TvfsFileStream.SetSize(NewSize: LongInt);
begin
  // Set the size property of the wrapped stream
  FSize:= NewSize;
  FFileRecord.FileSize := FSize;

  //TODO: can this be removed here and only be used in write?
  TVirtualFileSystem(self.FDataFileP).ChangeFile(FFileRecord.filename, FFileRecord);
end;

function TvfsFileStream.Read(var Buffer; Count: Integer): Longint;
begin
  result := TVirtualFileSystem(FDataFileP).ReadVirtualFileStreamBuffer(Buffer, FPosition, Count, FBlockList);
  FPosition:=FPosition+result; //update fileposition
end;

function TvfsFileStream.Seek(Offset: Integer; Origin: Word): Longint;
var
  len : integer;
begin
  len := FFileRecord.filesize;
  case Origin of
    soFromBeginning: FPosition := Offset;
    soFromCurrent: FPosition := FPosition + Offset;
    soFromEnd: FPosition := len - Offset;
  end;
  if (FPosition > len) then
  begin
    FPosition := len;
  end else
  if (FPosition < 0) then
  begin
    FPosition := 0;
  end;
  Result := FPosition;
end;

function TvfsFileStream.Write(const Buffer; Count: Integer): Longint;
var
  TempPosition: Int64;
begin
  //Determine if size has changed
  TempPosition:=FPosition; //save current fileposition
  if FPosition + Count > FSize then
  begin
    FSize := FPosition + Count;
    self.SetSize(FSize);
  end;
  FPosition:=TempPosition; //restore current filepostion

  //write the buffer
  result := TVirtualFileSystem(FDataFileP).WriteVirtualFileStreamBuffer(Buffer, FPosition, Count, FBlockList);

  //check if filerecord needs to be update due to changed startblock
  if FBlockList[0] <> FFirstBlock then
  begin
    ffilerecord.startblock:=FBlockList[0];
    TVirtualFileSystem(FDataFileP).ChangeFile(FFileRecord.filename,ffilerecord);
  end;

  if result = -1 then
  begin
    Raise Exception.Create('write returned wrong size');
  end;
  FPosition:=FPosition+result; //update fileposition
end;

//TvfsFileSystem ---------------------------------------------------------------

constructor TVirtualFileSystem.Create();
begin
  FDataFile := nil;
  FBlockSize := 512;//*1024;  //was 512
end;

destructor TVirtualFileSystem.Destroy();
begin
  if FDataFile <> nil then self.Close();
  inherited Destroy();
end;

procedure TVirtualFileSystem.New();
begin
  FDataFile:= TFileStream.Create(FDataFileName, fmCreate);
  self.WriteHeader(); //header file with virtualfilesystem info
  self.WriteFirstDirectoryEntry(); //first directory entry /
  self.Close();
  self.Open();
end;

procedure TVirtualFileSystem.New(aBlockSize: integer);
begin
  FBlockSize:=aBlockSize;
  self.New();
end;

procedure TVirtualFileSystem.Open();
begin
  FDataFile := TFileStream.Create(FDataFileName, fmOpenReadWrite);
  FDataFile.Position := 0;

  //open virtual directory filestream
  FDir := TvfsFileStream.Create('/',self);
  FDataFile.Position := 0;
end;

procedure TVirtualFileSystem.Close();
begin
  if Fdir <> nil then
  begin
    FDir.Free;
    FDir :=nil;
  end;
  FDataFile.Free;
  FDataFile := nil;
end;

function TVirtualFileSystem.ReadBlock(ablockid: int64; pBuf: pointer): int64;
var
  block: TBlock;
//  i: integer;
begin
  //TODO: as block is always read completely this can be simplified
  //remove the TBuffer copying?

//  GetMem(block.data, FBlockSize - (sizeof(int64)*2) );

  FDataFile.Position :=(ablockid * FBlockSize);
  FDataFile.ReadBuffer(block.previous, sizeof(block.previous));
  FDataFile.ReadBuffer(block.next, sizeof(block.next));
  FDataFile.ReadBuffer(pbuf^, FBlockSize - (sizeof(int64)*2) );

  //how to detect if block is not filled completely? Need to know filesize in ReadData function
//  for i:=0 to FBlockSize-1- (sizeof(int64)*2) do
//  begin
//    TBuffer(pBuf)[i]:=TBuffer(block.data)[i];
//  end;
//  FreeMem(block.data);

  result := block.next;
end;

function TVirtualFileSystem.ReadBlock(ablockid: int64; pBuf: pointer; beginpos: integer; endpos: integer): int64;
var
  block: TBlock;
//  i: integer;
begin
  //TODO: as block is always read completely this can be simplified
  //remove the TBuffer copying?

//  GetMem(block.data, FBlockSize - (sizeof(int64)*2) );

  FDataFile.Position :=(ablockid * FBlockSize);
  FDataFile.ReadBuffer(block.previous, sizeof(block.previous));
  FDataFile.ReadBuffer(block.next, sizeof(block.next));

  FDataFile.Position :=FDataFile.Position+beginpos;
  FDataFile.ReadBuffer(pbuf^, (endpos-beginpos) );

  //how to detect if block is not filled completely? Need to know filesize in ReadData function
//  for i:=0 to FBlockSize-1- (sizeof(int64)*2) do
//  begin
//    TBuffer(pBuf)[i]:=TBuffer(block.data)[i];
//  end;
//  FreeMem(block.data);

  result := block.next;
end;

function TVirtualFileSystem.ReadBlock(ablockid: int64; var buffer; beginpos: integer; endpos: integer): int64;
var
  block: TBlock;
//  i: integer;
begin
  //TODO: as block is always read completely this can be simplified
  //remove the TBuffer copying?

//  GetMem(block.data, FBlockSize - (sizeof(int64)*2) );

  FDataFile.Position :=(ablockid * FBlockSize);
  FDataFile.ReadBuffer(block.previous, sizeof(block.previous));
  FDataFile.ReadBuffer(block.next, sizeof(block.next));

  FDataFile.Position :=FDataFile.Position+beginpos;
  FDataFile.ReadBuffer(buffer, (endpos-beginpos) );

  //how to detect if block is not filled completely? Need to know filesize in ReadData function
//  for i:=0 to FBlockSize-1- (sizeof(int64)*2) do
//  begin
//    TBuffer(pBuf)[i]:=TBuffer(block.data)[i];
//  end;
//  FreeMem(block.data);

  result := block.next;
end;

procedure TVirtualFileSystem.WriteBlock(ablockid, anextblockid, aprevblockid: int64; pBuf: Pointer);
//var
//  Block: TBlock;
//  i: integer;
begin
  //TODO: as block is always read completely this can be simplyfied

//  block.previous := aprevblockid;
//  block.next := anextblockid;

//  i:=FBlockSize-(sizeof(int64)*2);
//  system.GetMem(block.data, i );
//  FillChar(block.data^, FBlockSize-(sizeof(int64)*2) , 0);

//  for i:=0 to (FBlockSize-(sizeof(int64)*2))-1 do
//  begin
//    TBuffer(block.data)[i]:=TBuffer(pBuf)[i];
//  end;


  FDataFile.Position := (ablockid * FBlockSize);

//  FDataFile.WriteBuffer(block.previous,sizeof(block.previous));
//  FDataFile.WriteBuffer(block.next,sizeof(block.next));

  FDataFile.WriteBuffer(aprevblockid,sizeof(aprevblockid));
  FDataFile.WriteBuffer(anextblockid,sizeof(anextblockid));

  FDataFile.WriteBuffer(pbuf^,FBlockSize-(sizeof(int64)*2) );

//  system.freemem(block.data);
//  block.data := nil;
end;

procedure TVirtualFileSystem.WriteBlock(ablockid, anextblockid, aprevblockid: int64; pBuf: Pointer; beginpos: integer; endpos: integer);
var
  Block: TBlock;
//  i: integer;
begin
  //TODO: as block is always read completely this can be simplyfied

  block.previous := aprevblockid;
  block.next := anextblockid;

//  i:=FBlockSize-(sizeof(int64)*2);
//  system.GetMem(block.data, i );
//  FillChar(block.data^, FBlockSize-(sizeof(int64)*2) , 0);

//  for i:=beginpos to endpos do
//  begin
//    TBuffer(block.data)[i]:=TBuffer(pBuf)[i];
//  end;

  FDataFile.Position := (ablockid * FBlockSize);

FDataFile.WriteBuffer(block.previous,sizeof(block.previous));
  FDataFile.WriteBuffer(block.next,sizeof(block.next));


    FDataFile.Position := FDataFile.Position + beginpos;


  //FDataFile.WriteBuffer(pbuf^,FBlockSize-(sizeof(int64)*2)-(endpos-beginpos) );
  FDataFile.WriteBuffer(pbuf^,(endpos-beginpos) );

//  system.freemem(block.data);
//  block.data := nil;
end;

procedure TVirtualFileSystem.WriteBlock(ablockid, anextblockid, aprevblockid: int64; const buffer; beginpos: integer; endpos: integer);
var
  Block: TBlock;
//  i: integer;
begin
  //TODO: as block is always read completely this can be simplyfied

  block.previous := aprevblockid;
  block.next := anextblockid;

//  i:=FBlockSize-(sizeof(int64)*2);
//  system.GetMem(block.data, i );
//  FillChar(block.data^, FBlockSize-(sizeof(int64)*2) , 0);

//  for i:=beginpos to endpos do
//  begin
//    TBuffer(block.data)[i]:=TBuffer(pBuf)[i];
//  end;

  FDataFile.Position := (ablockid * FBlockSize);

  FDataFile.WriteBuffer(block.previous,sizeof(block.previous));
  FDataFile.WriteBuffer(block.next,sizeof(block.next));

    FDataFile.Position := FDataFile.Position + beginpos;

  //FDataFile.WriteBuffer(pbuf^,FBlockSize-(sizeof(int64)*2)-(endpos-beginpos) );
  FDataFile.WriteBuffer(buffer,(endpos-beginpos) );

//  system.freemem(block.data);
//  block.data := nil;
end;

//------------------------------------------------------------------------------

function TVirtualFileSystem.ReadData(aname: string; pBuf: pointer): integer;
var
  size, count: integer;
  prevblock, nextblock: int64;
  buffer: pointer;
  i: integer;
  FileRecord: TFileRecord;
  FileRecordPos: int64;
begin
  //Determine if aname exists
  FileRecordPos := findfile(aname,FileRecord);
  if FileRecord.startblock > 0 then
  begin
    system.getmem(buffer, FBlockSize - (sizeof(int64)*2) );
    //Determine first block to read for aname
    nextblock := FileRecord.startblock;
    prevblock := 0;
    size := 0;
    while nextblock <> 0 do
    begin
      //count := FBlockSize -(sizeof(int64)*2);
      nextblock := ReadBlock(nextblock, buffer);
      for i:=0 to FBlockSize -1-(sizeof(int64)*2) do
      begin
        if i+size < Filerecord.filesize then
          TBuffer(pBuf)[i+size]:=TBuffer(buffer)[i];
      end;
      size := size + FBlockSize - (sizeof(int64)*2);
    end;
    system.freemem(buffer);
  end;
  result := FileRecord.filesize;
end;

procedure TVirtualFileSystem.WriteData(aname: string; pBuf: pointer; aSize: integer);
var
  buffer: pointer;
  i: integer;
  tempsize: integer;
  prevblock, nextblock: int64;
  blockcount: integer;
  FileRecord: TFileRecord;
  FileRecordPos: int64;
  FirstBlock: int64;
  FixedFirstBlock: int64;
begin
  FirstBlock:= 0;
  prevblock := 0;

  //Does aname exist?
  FileRecordPos := findfile(aname, filerecord);
  if filerecord.startblock > 0 then
  begin
    //Overwrite current file? Is this needed?
  end
  else
  begin
    //aname does not exist so create it
    FirstBlock:=self.NextFreeBlock;
    AddFile(aname,self.NextFreeBlock() , aSize);
  end;

  //check if the firstblock has changed due to directory entry creation
  if firstblock > 0 then
  begin
    fixedfirstblock:=self.NextFreeBlock();
    if fixedfirstblock <> firstblock then
    begin
      fillchar(filerecord, sizeof(TFileRecord), 0);
      filerecord.filename:=aname;
      filerecord.filesize:=asize;
      filerecord.startblock:=fixedfirstblock;
      self.ChangeFile(aname,filerecord);
    end;
  end;

  //if buffer is larger then blocksize then break it up in parts
  if aSize > (FBlockSize - (sizeof(int64)*2) )then
  begin
    tempsize := aSize;
    blockcount := 0;

    while tempsize > (FBlockSize - (sizeof(int64)*2) ) do
    begin
      system.GetMem(buffer, FBlockSize - (sizeof(int64)*2) );
      fillchar(buffer^, FBlockSize - (sizeof(int64)*2) , 0); //hier evt overbodig...
      for i:=0 to FBlockSize-1-(sizeof(int64)*2) do
      begin
        TBuffer(buffer)[i] := TBuffer(pBuf)[i+blockcount];
      end;
      nextblock := self.NextFreeBlock();
      self.WriteBlock(nextblock,self.SecondNextFreeBlock(),prevblock,buffer);
      prevblock := nextblock;
      tempsize := tempsize - FBlockSize + (sizeof(int64)*2);
      system.FreeMem(buffer);

      blockcount := blockcount + FBlockSize-(sizeof(int64)*2);//16;
    end;

    //fix to use correct last buffer part
    system.GetMem(buffer, FBlockSize-(sizeof(int64)*2));
    fillchar(buffer^, FBlockSize-(sizeof(int64)*2), 0);
    for i:=0 to tempsize-1 do
    begin
      TBuffer(buffer)[i] := TBuffer(pBuf)[i+blockcount];
    end;
    self.WriteBlock(self.NextFreeBlock(),0,prevblock,buffer);
    system.FreeMem(buffer);
  end
  else
  begin
    //just write the complete buffer if it is smaller then the blocksize
    system.GetMem(buffer, FBlockSize-(sizeof(int64)*2));
    fillchar(buffer^, FBlockSize-(sizeof(int64)*2), 0);
    for i:=0 to aSize-1 do
    begin
      TBuffer(buffer)[i] := TBuffer(pBuf)[i];
    end;
    self.WriteBlock(self.NextFreeBlock(),0,prevblock,buffer);
    system.FreeMem(buffer);
  end;
end;

//------------------------------------------------------------------------------

function TVirtualFileSystem.NextFreeBlock(): int64;
begin
  //Determine next free block (filesize / FBlockSize);
  //TODO: also check list of deleted/free block to reuse those
  //should i also make a special case for overwriting the same file?
  result := (FDataFile.Size div FBlockSize);
end;

function TVirtualFileSystem.SecondNextFreeBlock(): int64;
begin
  //Determine the second next free block (filesize / FBlockSize);
  //TODO: also check list of deleted/free block to reuse those
  //should i also make a special case for overwriting the same file?
  result := (FDataFile.Size div FBlockSize)+1;
end;

procedure TVirtualFileSystem.WriteFirstDirectoryEntry();
var
  FileRecord : TFileRecord;
  pBuf : pointer;
//  i: integer;
begin
  fillchar(filerecord, SizeOf(TFileRecord),0); //clean up memory
  FileRecord.filename := '/';
  FileRecord.startblock := self.NextFreeBlock();
  FileRecord.filesize := SizeOf(TFileRecord);

  getMem(pBuf, FBlockSize-(sizeof(int64)*2));
  FillChar(pBuf^, FBlockSize-(sizeof(int64)*2) , 0);

//  for i:=0 to ( SizeOf(TFileRecord) )-1 do
//  begin
//    TBuffer(pbuf)[i]:=TBuffer(@filerecord)[i];
//  end;

  self.WriteBlock(1,0,0,pbuf);
  freemem(pbuf);
  //writes first directory entry
  self.WriteBlock(1,0,0,@FileRecord, 0, SizeOf(TFileRecord));

end;

procedure TVirtualFileSystem.WriteHeader();
var
  FileRecord : TFileRecord;
  pBuf: pointer;
begin
  fillchar(filerecord, SizeOf(TFileRecord),0); //clean up memory
  FileRecord.filename := 'Header Data';
  FileRecord.startblock := 0;

  getMem(pBuf, FBlockSize-(sizeof(int64)*2));
  FillChar(pBuf^, FBlockSize-(sizeof(int64)*2) , 0);
  self.WriteBlock(0,0,0,pbuf);
  freemem(pbuf);
  self.WriteBlock(0,0,0,@FileRecord, 0, SizeOf(TFileRecord));
end;

//------------------------------------------------------------------------------

function  TVirtualFileSystem.ReadStream(Section, Ident: string; Stream: TStream): Integer;
var
  buffer: pointer;
  size: int64;
  FileRecord: TFileRecord;
  FileRecordPos: int64;
begin
  FileRecordPos := findfile(Ident,FileRecord);
  system.getmem(buffer, FileRecord.filesize);
  size:=self.ReadData(Ident, buffer);
  stream.Write(buffer^, size);
  system.freemem(buffer);
  result := FileRecord.filesize;
end;

function  TVirtualFileSystem.WriteStream(Section, Ident: string; Stream: TStream): Integer;
var
  buffer: pointer;
begin
  system.GetMem(buffer,stream.Size);
  stream.Read(buffer^,stream.Size);
  self.WriteData(Ident, buffer, stream.Size);
  system.FreeMem(buffer);
  result := stream.Size;
end;

//------------------------------------------------------------------------------

procedure TVirtualFileSystem.GetFileList(List: TStrings);
var
  i,m: Integer;
  FileRecordEntry: TFileRecord;
  OrigFilePos : int64;
begin
  OrigFilePos:= FDir.FPosition;
  FDir.FPosition :=0;
  i:=0;
  m:= (FDir.Size div SizeOf(TFileRecord)) ;
  while i<=m-1 do
  begin
    FileRecordEntry.filename := 'Not Found';
    FileRecordEntry.filesize := 0;
    FileRecordEntry.startblock := 0;
    FDir.Read(FileRecordEntry, SizeOf(TFileRecord) );
    List.Add(FileRecordEntry.filename+'-'+IntToStr(FileRecordEntry.filesize)+'-'+IntToStr(FileRecordEntry.startblock) );
    i:=i+1;
  end;
  FDir.FPosition :=OrigFilePos;
end;

function TVirtualFileSystem.FindFile(aname: string; var afilerecord: TFileRecord): int64;
var
  FileRecordEntry: TFileRecord;
  buffer: pointer;
  i,m,size: integer;
  foundpos: integer;
  OrigFilePos: int64;
begin
  result :=0;
  FileRecordEntry.filename := 'Not Found';
  FileRecordEntry.filesize := 0;
  FileRecordEntry.startblock := 0;
  afilerecord :=FileRecordEntry;
  foundpos:=0;
  if aname = '/' then
  begin
    //read first dir entry
    //TODO: can this be simplified by assuming a fixed /
//    system.GetMem(buffer, FBlockSize-(sizeof(int64)*2) );
    //size:= FBlockSize-(sizeof(int64)*2);
//    self.ReadBlock(1,buffer);

    self.ReadBlock(1,FileRecordEntry,0,SizeOf(TFileRecord));
//    for i:=0 to sizeof(tfilerecord) do
//    begin
//      TBuffer(@FileRecordEntry)[i]:=TBuffer(buffer)[i];
//    end;
//    system.FreeMem(buffer);

    //compare
    if FileRecordEntry.filename = aname then
    begin
      result := foundpos * SizeOf(TFileRecord);
      afilerecord := FileRecordEntry;
    end;
    //foundpos := foundpos + 1;
  end
  else
  begin
    //browse through FileRecord in virtual file stream
    OrigFilePos:=FDir.FPosition;
    FDir.FPosition :=0;
    i:=0;
    m:= (FDir.Size div SizeOf(TFileRecord));
    while i<=m-1 do
    begin
      FDir.Read(FileRecordEntry, SizeOf(TFileRecord) );
      //compare
      if FileRecordEntry.filename = aname then
      begin
        result := FDir.FPosition-SizeOf(TFileRecord) ;
        afilerecord := FileRecordEntry;
        FDir.FPosition :=OrigFilePos;
        break;
      end;
      i:=i+1;
    end;
    FDir.FPosition :=OrigFilePos;
  end;
end;

procedure TVirtualFileSystem.AddFile(aname: string; astartblockid: int64; aSize: int64);
var
  FileRecord: TFileRecord;
begin
  FDir.Position := FDir.Size; //go to end of virtual file stream
  fillchar(filerecord, SizeOf(TFileRecord),0); //clean up memory
  FileRecord.filename := aname;
  FileRecord.startblock := astartblockid;
  FileRecord.filesize := aSize;
  FDir.Write(FileRecord, SizeOf(TFileRecord));
end;

procedure TVirtualFileSystem.ChangeFile(aname: string; afilerecord: TFileRecord);
var
  OrigFileRecord: TFileRecord;
  NewFileRecord: TFileRecord;
  OrigFileRecordPos: int64;
begin
  FillChar(NewFileRecord, SizeOf(TFileRecord),0);
  NewFileRecord.filename := aFileRecord.filename;
  NewFileRecord.filesize := aFileRecord.filesize;
  NewFileRecord.startblock := aFileRecord.startblock;

  OrigFileRecordPos := self.FindFile(aname,OrigFileRecord);
  if OrigFileRecord.filename <> 'Not Found' then
  begin
    if OrigFileRecordPos > 0 then
      FDir.FPosition := OrigFileRecordPos
    else
      FDir.FPosition := 0;

    FDir.Write(NewFileRecord, SizeOf(TFileRecord) );
  end;
end;

//------------------------------------------------------------------------------

function TVirtualFileSystem.GetBlockList(afile: TFileRecord): TBlockList;
var
  i: integer;
//  totalblocks: integer;
  p, n: int64;
  SIZE: integer;
  FILESIZE: integer;
begin
  FILESIZE:=afile.filesize;
  SIZE:= FBlockSize - (2*SizeOf(int64) );

  //totalblocks calculation is wrong?
  //totalblocks := (FileSize div SIZE)+2; //why do i have to add 2?

  if afile.filename = 'Not Found' then
    result := nil
  else
  begin
    SetLength(result, 1);
    n:=afile.startblock;
    result[0] := n;
  end;

  i:=1;

  if FILESIZE > SIZE then
  begin
    while n <> 0 do
    begin
      SetLength(result, i+1);
      FdataFile.Position := n * FBlockSize ;
      FDataFile.ReadBuffer(p, sizeof(p));
      FDataFile.ReadBuffer(n, sizeof(n));
      result[i] := n;
      i:=i+1;
    end;
  end;
end;

//------------------------------------------------------------------------------

//TODO: Optimize code below
//TODO: use special ReadBlock/Writeblock for writing reading only part of block

function TVirtualFileSystem.ReadVirtualFileStreamBuffer(var buffer; Position, Count: Integer; BlockList: TBlockList): LongInt;
var
   ibegin, iend: integer;
   iBeginBlock, iEndBlock: integer;
   SIZE: integer;
   i: integer;
   mySize: integer;
   testteller: integer;
begin
  result := count;

  ibegin := Position; //beginposition in virtual filestream
  iEnd   := Position+count-1; //endposition in virtual filestream

  SIZE:= FBlockSize - (2*SizeOf(int64) ); //blocksize corrected with prev next

  iBeginBlock := iBegin div Size; //block id for beginposition
  iEndBlock := iEnd div Size; //block id for endposition

  mySize:=0; //read bytes (goes up by blocksize)

  //if position is not in the first block of virtual file then correct ibegin and iend
  if iBeginBlock > 0 then
  begin
      iBegin := iBegin - (iBeginBlock*SIZE);
      iEnd := iEnd - (iBeginBlock*SIZE);
  end;

  testteller:=0; //read bytes
  mysize:=iEnd;

  //if we need to read from multiple blocks
  if iBeginBlock <> iEndBlock then
  begin

    for i:=iBeginBlock to iEndBlock do
    begin
      //first block
      if i=iBeginBlock then
      begin
        self.ReadBlock(BlockList[i], Addr(TBuffer(@buffer)[0]), iBegin,  SIZE);
        testteller := SIZE-iBegin;
        mysize := mysize - (SIZE-iBegin);
      end;

      //middle blocks read complete
      if (i>iBeginblock) and (i<iEndblock) then
      begin
        self.ReadBlock(BlockList[i], Addr(TBuffer(@buffer)[testteller]) );
        testteller := testteller+SIZE;
        MySize := MySize - SIZE;
      end;

      //last block
      if i=iEndBlock then
      begin
        self.ReadBlock(BlockList[i], Addr(TBuffer(@buffer)[testteller]), 0, mySize-iBegin+1 );
        testteller := testteller + (mysize-iBegin) +1; 
      end;
    end;
  end
  else
  begin
    //smaller then count so read only first block
    self.ReadBlock(BlockList[iBeginBlock], buffer, ibegin, iend+1); //waarom +1 ?
    testteller := count;
 end;

 //check if we read the expected amount of bytes
 if testteller<> count then
  result := -1
 else
  result:=testteller;

end;

function TVirtualFileSystem.WriteVirtualFileStreamBuffer(const buffer; Position, Count: Integer; var BlockList: TBlockList ): LongInt;
var
  writebuffer: pointer;
  i,t: integer;
  tempsize: integer;
  prevblock, nextblock: int64;
  blockcount: integer;
  ibegin, iend: integer;
  iBeginBlock, iEndBlock: integer;
  Size, mysize: integer;
  b,nb,tnb: int64;
  blocklistsize: integer;
  tempbuffer: pointer;
  testteller: int64;
  existingblock: boolean;

begin
  existingblock:= false;

  tempbuffer := @Buffer; //pointer to buffer

  result := count;

  ibegin := Position; //beginposition in virtual filestream
  iEnd   := Position+count;//-1; //beginposition in virtual filestream

  SIZE:= FBlockSize - (2*SizeOf(int64) ); //blocksize corrected with prev next

  iBeginBlock := iBegin div Size; //blockid for first block
  iEndBlock := iEnd div Size; //blockid for last block

  blockListSize:=High(BlockList)+1; //get the number of blocks

  system.GetMem(writebuffer, SIZE); //make a writebuffer

  testteller := 0; //read bytes

  mySize:=0; //read bytes (goes up by blocksize)


  //if position is not in the first block of virtual file then correct ibegin and iend
  if iBeginBlock > 0 then
  begin
    iBegin := iBegin - (iBeginBlock*SIZE);
    iEnd := iEnd - (iBeginBlock*SIZE);
  end;

   mySize:=iEnd;

  //if we need to read from multiple blocks
  if iBeginBlock <> iEndBlock then
  begin
    //for each block from begin to endblock
    for i:=iBeginBlock to iEndBlock do
    begin
      existingblock := false;
      b:=0;
      nb:=0;
      GetBlockId(existingblock, iEndBlock, blocklistsize, BlockList, i, b, nb);

      //first block
      if i=iBeginBlock then
      begin
        if existingblock = false then
        begin
          fillchar(writebuffer^, SIZE, 0); //clean the write buffer
          self.WriteBlock(b,nb,0,writebuffer);
        end;
        self.WriteBlock(b,nb,0, Addr(TBuffer(@buffer)[0]), iBegin,  SIZE);
        testteller := SIZE-iBegin;
        mysize := mysize - (SIZE-iBegin);
      end;

      //middle blocks read complete
      if (i>iBeginblock) and (i<iEndblock) then
      begin
        self.WriteBlock(b,nb,0, Addr(TBuffer(@buffer)[testteller]) );
        testteller := testteller+SIZE;
        MySize := MySize - SIZE;
      end;

      //last block
      if i=iEndBlock then
      begin
        if existingblock = false then
        begin
          fillchar(writebuffer^, SIZE, 0); //clean the write buffer
          self.WriteBlock(b,nb,0,writebuffer);
        end;
        self.WriteBlock(b,nb,0, Addr(TBuffer(@buffer)[testteller]), 0, mySize-iBegin );
        testteller := testteller + (mysize-iBegin) ;
      end;
    end;
  end
  else
  begin
    existingblock := false;
    //determine what block to use
    b:=0;
    nb:=0;
    GetBlockId(existingblock, iEndBlock, blocklistsize, BlockList, iBeginBlock, b, nb);
    if existingblock = false then
    begin
      fillchar(writebuffer^, SIZE, 0); //clean the write buffer
      self.WriteBlock(b,nb,0,writebuffer); //a new block needs to be written completely
    end;

    self.WriteBlock(b, nb, 0, buffer, ibegin, iend);
    testteller := count;
  end;

  system.FreeMem(writebuffer); //clean up the write buffer

  //check if we read the expected amount of bytes
  if testteller<> count then
    result := -1
  else
    result:=testteller;

end;

procedure TVirtualFileSystem.GetBlockId(var existingblock: Boolean; iEndBlock: Integer; blocklistsize: Integer; var BlockList: TBlockList; i: Integer; var b: Int64; var nb: Int64);
var
  tnb: Int64;
begin
  //determine what block to use or create
  if i > High(BlockList) then
  begin
    b := self.NextFreeBlock;
    nb := self.SecondNextFreeBlock;
    if BlockList = nil then
    begin
      BlockListSize := 0;
      BlockListSize := BlockListSize + 1;
      SetLength(BlockList, BlockListSize);
    end
    else
    begin
      BlockListSize := BlockListSize + 1;
      SetLength(BlockList, BlockListSize);
    end;
    BlockList[i] := b;
  end
  else
  //clean the write buffer
  begin
//    if BlockList[i] = 0 then
//    begin
//      //als i gelijk aan blocklist met waarde 0 dan nieuw block maken
//      b:=self.NextFreeBlock();
      //nb:=self.SecondNextFreeBlock();
//     nb:=0;
//      BlockList[i]:=b;
//    end;
    tnb := 0;
    b := BlockList[i];
    if (i + 1) > High(BlockList) then
    begin
      tnb := self.NextFreeBlock;
    end
    else
    begin
      if (i + 1 > High(BlockList)) then
        nb := 0
      else
        nb := BlockList[i + 1];
    end;

    existingblock := true;
    if tnb > 0 then
      nb := tnb;

  end;
  if i >= High(BlockList) then
    //iEndBlock then correct nb to 0
    if i = iEndBlock then
      nb := 0;

end;

end.
